/*
Validating processor of an XML Regular Schema Language (Ver 0.60)
Copyright (C) 1999      MURATA Makoto (FAMILY Given)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

MURATA Makoto can be reached at makoto@mars.netspace.or.jp.
If this address does not work, use murata.makoto@fujixerox.co.jp.

Contributor(s): 	

*/

//////////////////////////////////////////////////////////////////////
//
// InstanceHandler.cpp: Implementation of the class InstanceHandler
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Handlers for the start/end of a document
//////////////////////////////////////////////////////////////////////

void
InstanceHandler::startDocument()
{
	//clear text
	DOMString empty(TEXTBUFFERLENGTH);
	currentText = empty;

	current_fm_states = new FmStates* [2];
	current_fm_states[0] = new FmStates(&(schema_p->top_fm));
	current_fm_states[1] = 0;
}

void
InstanceHandler::endDocument()
{

	if (validity) {
		//process text
		concatenatedCharacter();
		
		set<state> fins;
		current_fm_states[0]->fm_p->finals(fins);
		set<state> intersection;
		intersection.intersect(fins, current_fm_states[0]->currents);
		if (intersection.size() != 0)
			validity = true;
		else 
			validity = false;
	}
}

//////////////////////////////////////////////////////////////////////
// Handlers for the start/end of each element
//////////////////////////////////////////////////////////////////////

void
InstanceHandler::startElement(const XMLCh* const name_p, 
							AttributeList& attributes) 
{	
	if (debug_option)
		outStrm << "startElement" << EndLn;

	//process text
	concatenatedCharacter();
	//clear text
	DOMString empty(TEXTBUFFERLENGTH);
	currentText = empty;

	//simulation of non-deterministic bottom-up hedge automata
	push (current_fm_states);
	current_fm_states = new FmStates* [MAXRULESPERTAGTYPE];
	
	//In the current implementation, we assume that there is 
	//one to one correspondence between tagTypes and "element types"
	//of XML 1.0.  Attribute validation is not implemented.

	RulesPerTagType* rule_list_p = 
		schema_p->ruleHashtable.get(name_p);
	if (!rule_list_p) {
		validity = false;
		errStrm << "undefined tagType: " << name_p << EndLn;
		throw SAXParseException(
			XMLString::transcode("undefined tagType"), 
			*locator_p);
	}
	for (int i = 0; rule_list_p->rule_p_p[i] != 0; i++) {
			Rule* each_rule_p = rule_list_p->rule_p_p[i];
			current_fm_states[i] = 
				new FmStates(&(each_rule_p->rhs_fm), &(each_rule_p->nt));
			if (debug_option) 
				outStrm << "current FmStates: " << EndLn
				<< *(current_fm_states[i]);
	}
	current_fm_states[i] = 0;
}

void
InstanceHandler::endElement(const XMLCh* const name_p)
{
	if (debug_option)
		outStrm << "endElement" << EndLn;
	
	//process text
	concatenatedCharacter();
	//clear text
	DOMString empty(TEXTBUFFERLENGTH);
	currentText = empty;

	//simulation of non-deterministic bottom-up hedge automata
	set<NonTerm> current_nonTerms;
	for (int i = 0; current_fm_states[i] != 0; i++) {
		set<state> fins;
		current_fm_states[i]->fm_p->finals(fins);
		set<state> intersection;
		intersection.intersect(fins, current_fm_states[i]->currents);
		if (intersection.size() != 0)
			current_nonTerms += *(current_fm_states[i]->nt_p);
		if (debug_option)
			outStrm << "current FmStates: " << *(current_fm_states[i]);
	}
	if (debug_option) {
		outStrm << "endElement" << EndLn;
		outStrm << "computed NonTerms: " << current_nonTerms;
	}
	if (current_nonTerms.size() == 0) {
		validity = false;
		errStrm << "illeagal subordinates: " << name_p << EndLn;
		throw SAXParseException(
			XMLString::transcode("illeagal subordinates"), 
			*locator_p);
	}

	current_fm_states = pop();
	updateFmStates(current_nonTerms);
}

//////////////////////////////////////////////////////////////////////
// Handlers for character data.  
// Note: unlike expat, XML4C concatenates every line.
//////////////////////////////////////////////////////////////////////

void
InstanceHandler::characters(const XMLCh* const chars, 
							const unsigned int length)
{
	DOMString another(chars);

	currentText.appendData(another);
}

//////////////////////////////////////////////////////////////////////
// Compute non-terminals from the concatenated character data.
//////////////////////////////////////////////////////////////////////

void InstanceHandler::concatenatedCharacter()
{
	set<NonTerm> current_nonTerms;

	RefHashTableOfEnumerator<Datatype> 
		datatype_enum(&(schema_p->datatypeHashtable));

	XMLCh* chars = currentText.rawBuffer();

	while (datatype_enum.hasMoreElements()) {
		Datatype* dt_p = &(datatype_enum.nextElement());
		if ((*(dt_p->fun_p))(chars)) {
			current_nonTerms += 
				dt_p->getNonTerm();
		}
	}

	if (debug_option) {
		outStrm << "characters" << EndLn;
		outStrm << "computed NonTerms: " << current_nonTerms;
	}
	updateFmStates(current_nonTerms);
}

//////////////////////////////////////////////////////////////////////
// Update current states of each of the current automata.
//////////////////////////////////////////////////////////////////////

void
InstanceHandler::updateFmStates(const set<NonTerm>& current_nonTerms) 
{
	if (debug_option)
		outStrm << "updateFmStates" << EndLn;

	bool found = false;

	for (int k = 0; current_fm_states[k] != 0; k++) {
		set<state> next_states;
		for (int i = 0; i < current_fm_states[k] -> currents.size(); i++) {
			state s = current_fm_states[k] -> currents[i];
			for (int j = 0; j < current_nonTerms.size(); j++) {
				if (current_nonTerms[j] == NonTerm::whitespace) {
					next_states += s;
					found = true;
					continue;
				}
				fm<NonTerm>	temp_fm;
				current_fm_states[k]
					->fm_p
						->select(current_nonTerms[j], s, SOURCE, temp_fm);
				if (temp_fm.size() == 0 && debug_option)
			      outStrm << "no states acccessible on " <<
							 current_nonTerms[j] << EndLn;
				set<state> temp_sts;
				temp_fm.sinks(temp_sts);
				for (int l = 0; l < temp_sts.size(); l++) {
					found = true;
					next_states += temp_sts[l];
				}
			}
		}
		if (debug_option) {
			outStrm << "current states: " << current_fm_states[k] -> currents;
			outStrm << "nonTerms: " << current_nonTerms;
		}
		current_fm_states[k] -> currents = next_states;
		if (debug_option)
			outStrm << "updated states: " << current_fm_states[k] -> currents;
	}
	if (!found) {
		validity = false;
		errStrm << "unexpected non-terminal: " << current_nonTerms << EndLn;
		throw SAXParseException(
			XMLString::transcode("unexpected non-terminal"), 
			*locator_p);
	}
}

//////////////////////////////////////////////////////////////////////
// Obtain a locator
//////////////////////////////////////////////////////////////////////

void InstanceHandler::setDocumentLocator(Locator *const locator)
{
	locator_p = locator;
}

