/*
Validating processor of an XML Regular Schema Language (Ver 0.60)
Copyright (C) 1999      MURATA Makoto (FAMILY Given)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

MURATA Makoto can be reached at makoto@mars.netspace.or.jp.
If this address does not work, use murata.makoto@fujixerox.co.jp.

Contributor(s): 	

*/

//////////////////////////////////////////////////////////////////////
//
// Schema.cpp: Implementation of the class Schema
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// constructors/destructors
//////////////////////////////////////////////////////////////////////

Schema::Schema(const DOM_Document& dom_doc, bool debug)
  : dom_schema(dom_doc),
	ruleHashtable(MAXTAGTYPES, true),
    nonTermHashtable(MAXNONTERMS, true),
	tagTypeLessNonTermHashtable(MAXTAGTYPELESSNONTERMS, true),
	datatypeHashtable(MAXDATATYPES, true),
	debug_option(debug),
	D_null(""),
	D_zero("0"),
	D_one("1"),
	D_infinity("*")
{
}

Schema::~Schema()
{

}

//////////////////////////////////////////////////////////////////////
// helper function
//////////////////////////////////////////////////////////////////////

static DOM_Element firstElementChild(const DOM_Element& elem_node)
{
	DOM_NodeList children = elem_node.getChildNodes();
	for (int i = 0; i < children.getLength(); i++)
		if (children.item(i).getNodeType() == DOM_Node::ELEMENT_NODE)
			break;
	if (i == children.getLength())
		throw XMLString::transcode("no element subordinates!");
	return (DOM_Element&)(children.item(i));
}

//////////////////////////////////////////////////////////////////////
//  Creating hash tables and automata
//////////////////////////////////////////////////////////////////////

void Schema::initialize()
{
	Datatype::initialize(*this);
	tagTypeLessRuleInitialize();
	ruleInitialize();
	topLevelInitialize();
}

void Schema::topLevelInitialize()
{
	DOM_NodeList topLevel_list = dom_schema.getElementsByTagName("topLevel");
	if (topLevel_list.getLength() == 0)
		throw XMLString::transcode("no topLevel!");
	DOM_Element toplevel_node = 
		(DOM_Element&) topLevel_list.item(0);
	DOM_Element top_cm = 
		firstElementChild(toplevel_node);
	cm2fm(top_cm, top_fm);
}

void Schema::ruleInitialize()
{
	DOM_NodeList rule_node_list =
		dom_schema.getElementsByTagName("rule");

	for (int i = 0; i < rule_node_list.getLength(); i++) {
		DOM_Element rule_node = (DOM_Element&)rule_node_list.item(i);
		DOMString tagTypeName = rule_node.getAttribute("tagType");
		if (tagTypeName.equals(""))
			continue;
		DOMString nonTermName = rule_node.getAttribute("nonTerm");
		if (nonTermName.equals("")) 
			 nonTermName =  tagTypeName;
		if (debug_option)
			outStrm << tagTypeName.rawBuffer() << EndLn <<
						nonTermName.rawBuffer() << EndLn;
		NonTerm nt(nonTermHashtable, nonTermName);
		Rule* rl_p = new Rule (nt, rule_node);
		DOM_Element cme_node = firstElementChild(rule_node);
		cm2fm(cme_node, rl_p->rhs_fm);

		RulesPerTagType* rule_list_p = 
			(RulesPerTagType*)ruleHashtable.get(tagTypeName.rawBuffer());
		if (!rule_list_p) {
			rule_list_p = new RulesPerTagType(tagTypeName);
			ruleHashtable.put(rule_list_p);
		}
		for (int j = 0; rule_list_p->rule_p_p[j] != 0; j++);
		rule_list_p->rule_p_p[j] = rl_p;
		rule_list_p->rule_p_p[j+1] = 0;
	}
}

void Schema::tagTypeLessRuleInitialize()
{
	DOM_NodeList rule_node_list =
		dom_schema.getElementsByTagName("rule");

	for (int i = 0; i < rule_node_list.getLength(); i++) {
		DOM_Element rule_node = (DOM_Element&)rule_node_list.item(i);
		DOMString tagTypeName = rule_node.getAttribute("tagType");
		if (!tagTypeName.equals(""))
			continue;
		DOMString nonTermName = rule_node.getAttribute("nonTerm");
		if (nonTermName.equals("")) 
			throw XMLString::transcode("The attribute \"nonTerm\" is not specified!");
		TagTypeLessNonTerm* mg_p = new TagTypeLessNonTerm (rule_node);
		tagTypeLessNonTermHashtable.put(mg_p);
	}

}
//////////////////////////////////////////////////////////////////////
// Creating automata from content models
//////////////////////////////////////////////////////////////////////

void
Schema::cm2fm(const DOM_Element& dom_elem, fm<NonTerm>& f)
{
	DOMString gi = dom_elem.getTagName();
	if (gi.equals("datatypeRef")) {
		cm2fmDatatypeRef(dom_elem, f);
	}
	else if (gi.equals("nonTermRef")) {
		cm2fmNonTermRef(dom_elem, f);
	}
	else if (gi.equals("empty")) {
		cm2fmEmpty(dom_elem, f);
	}
	else if (gi.equals("mixed")) {
		cm2fmMixed(dom_elem, f);
	}
	else if (gi.equals("sequence")) {
		cm2fmSequence(dom_elem, f);
	}
	else if (gi.equals("choice")) {
		cm2fmChoice(dom_elem, f);
	}
	else if (gi.equals("none")) {
		cm2fmNone(dom_elem, f);
	}
	else if (gi.equals("anyElement")) {
		cm2fmAnyElement(dom_elem, f);
	}
	else 
		_ASSERT(0);
}

void Schema::cm2fmDatatypeRef(const DOM_Element& dom_elem, fm<NonTerm>& f)
{
	DOMString name = dom_elem.getAttribute("name");
	NonTerm nt(datatypeHashtable, name);
	f.single(nt);
}

void Schema::cm2fmNonTermRef(const DOM_Element& dom_elem, fm<NonTerm>& f)
{
	DOMString name = dom_elem.getAttribute("name");
	const XMLCh* const name_p = name.rawBuffer();
	TagTypeLessNonTerm* mg_p = 
		tagTypeLessNonTermHashtable.get(name_p);
	if (mg_p)
		cm2fm(firstElementChild(mg_p->TagTypeLessNonTerm_elem), f);
	else {
		NonTerm nt (nonTermHashtable, name);
		f.single(nt);
	}
	minMax(dom_elem, f);
}

void Schema::cm2fmEmpty(const DOM_Element& dom_elem, fm<NonTerm>& f)
{
	f.empty_string_machine();
}

void Schema::cm2fmMixed(const DOM_Element& dom_elem, fm<NonTerm>& f)
{
	f.empty_string_machine();

	state max_st = f.max_state();
	inst<NonTerm> inst;
	fm<NonTerm> temp_fm;

	DOMString str("string");
	NonTerm nt(datatypeHashtable, str);
	inst.assign(max_st, nt, max_st);
	f.add_instruction(inst);

	DOM_NodeList dnl = dom_elem.getChildNodes();

	for (int i = 0; i < dnl.getLength (); i++) {
		if (dnl.item(i).getNodeType() != DOM_Node::ELEMENT_NODE)
			continue;
		cm2fm((DOM_Element&)dnl.item(i), temp_fm);
		set<NonTerm> labels;
		temp_fm.labels(labels);
		for (int j = 0; j < labels.size(); j++) {
			inst.assign(max_st, labels[j], max_st);
			f.add_instruction(inst);
		}
	}
}


void Schema::cm2fmSequence(const DOM_Element& dom_elem, fm<NonTerm>& f)
{
	f.empty_string_machine();
	DOM_NodeList dnl = dom_elem.getChildNodes();
	for (int i = 0; i < dnl.getLength (); i++) {
		if (dnl.item(i).getNodeType() != DOM_Node::ELEMENT_NODE)
			continue;
		fm<NonTerm> temp_fm;
		cm2fm((DOM_Element&)dnl.item(i), temp_fm);
		f ^= temp_fm;
	}
	minMax(dom_elem, f);
}

void Schema::cm2fmChoice(const DOM_Element& dom_elem, fm<NonTerm>& f)
{
	DOM_NodeList dnl = dom_elem.getChildNodes();
	if (cm2fmChoice_easy(dnl,f)) {
		minMax(dom_elem, f);
		return;
	}
	f.clear();
	for (int i = 0; i < dnl.getLength (); i++) {
		if (dnl.item(i).getNodeType() != DOM_Node::ELEMENT_NODE)
			continue;

		fm<NonTerm> temp_fm;
		cm2fm((DOM_Element&)dnl.item(i), temp_fm);

		f += temp_fm;
	}
	minMax(dom_elem, f);
}

bool Schema::cm2fmChoice_easy(const DOM_NodeList& dnl, fm<NonTerm>& f)
{
	f.clear();
	state start = 0;
	state final = 1;
	inst<NonTerm> inst;

	for (int i = 0; i < dnl.getLength (); i++) {
		if (dnl.item(i).getNodeType() == DOM_Node::ELEMENT_NODE) {
			DOM_Element each_sub_node = (DOM_Element&) dnl.item(i);
			DOMString gi = each_sub_node.getTagName();
			if (!(gi.equals("nonTermRef"))) {
				return false;
			}
			DOMString name = each_sub_node.getAttribute("name");
			if (name.equals(""))
				throw XMLString::transcode("The name attribute is missing!");
			NonTerm nt (nonTermHashtable, name);
			inst.assign(start, nt, final);
			f.add_instruction(inst);
		}
	}
	set<state> starts;
	starts += start;
	f.set_starts(starts);
	set<state> finals;
	finals += final;
	f.set_finals(finals);
	return true;
}

void Schema::cm2fmNone(const DOM_Element& dom_elem, fm<NonTerm>& f)
{
	f.clear();
}

void Schema::cm2fmAnyElement(const DOM_Element& dom_elem, fm<NonTerm>& f)
{
	throw XMLString::transcode("AnyElement is not supported yet.");
}

void Schema::minMax(const DOM_Element& dom_elem, fm<NonTerm> &f)
{

	DOMString minOccur = dom_elem.getAttribute("minOccur");
	if (minOccur.equals(D_null))
		minOccur = D_one;
	DOMString maxOccur = dom_elem.getAttribute("maxOccur");
	if (maxOccur.equals(D_null))
		maxOccur = D_one;

	
	if (minOccur.equals(D_zero) && maxOccur.equals(D_one)) {
		fm<NonTerm> temp_fm;
		temp_fm.empty_string_machine();
		f += temp_fm;
	}
	else if (minOccur.equals(D_zero) && maxOccur.equals(D_infinity)) {
		f.star();
	}
	else if (minOccur.equals(D_one) && maxOccur.equals(D_one)) {
	}
	else if (minOccur.equals(D_one) && maxOccur.equals(D_infinity)) {
		fm<NonTerm> temp_fm;
		temp_fm = f;
		f.star();
		f += temp_fm;
	}
	else
	  throw XMLString::transcode("Numeric minOccur or maxOccur is not supported.");
	f.subset();
	f.min_by_partition();
}


//////////////////////////////////////////////////////////////////////
// Stream output
//////////////////////////////////////////////////////////////////////

TextOutputStream&
operator<<(TextOutputStream& tos,  Schema& s)
{
	tos << "Top-level fm: " << EndLn << s.top_fm;

	RefHashTableOfEnumerator<RulesPerTagType> rule_enum(&(s.ruleHashtable));
	for (int i = 0; rule_enum.hasMoreElements(); i++)
		tos << rule_enum.nextElement();
	return tos;
}