/*
Validating processor of an XML Regular Schema Language (Ver 0.60)
Copyright (C) 1999      MURATA Makoto (FAMILY Given)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

MURATA Makoto can be reached at makoto@mars.netspace.or.jp.
If this address does not work, use murata.makoto@fujixerox.co.jp.

Contributor(s): 	

*/

//////////////////////////////////////////////////////////////////////
//
// regular3.cpp : the main program
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// standard output and error ouput
//////////////////////////////////////////////////////////////////////

XMLStdErr  errStrm;
XMLStdOut  outStrm;


//////////////////////////////////////////////////////////////////////
//  Local helper methods
//////////////////////////////////////////////////////////////////////
static void usage()
{
    outStrm <<  "\nUsage: rsl [options] schema instance\n"
            <<  "   This program validates an XML instance against a regular schema.\n"
			<<  "   This version (0.6) validates element hierarchies by executing\n"
			<<	"   hedge automata.  But it does not validate data against datatypes,\n"
			<<	"   and does not validate attributes.\n\n"
            <<  "   Options:\n"
            <<  "     -?          Show this help\n"
            <<  "     -V or -v    Validate schemata against the RSL DTD\n"
            <<  "     -S or -s    Debug option for schemata\n"
            <<  "     -I or -i    Debug option for instances\n"
            <<  EndLn;
}

//////////////////////////////////////////////////////////////////////
//  Validating instances
//////////////////////////////////////////////////////////////////////

static bool read_instance(Schema& schm, const char* instanceFile, 
						  bool debug_option)
{
	//
    //  Create a NonValidatingSAXParser and an InstanceHandler. 
    //

	NonValidatingSAXParser* inst_parser = new NonValidatingSAXParser;
    InstanceHandler*  inst_handler = new InstanceHandler(&schm, debug_option);
    inst_parser->setDocumentHandler(inst_handler);

	//
    //  Attach an error handler to the parser. The parser will call back to
    //  methods of the ErrorHandler if it discovers errors during the course
    //  of parsing the XML document.
    //

    ErrorHandler *errReporter = new ErrorReporter();
    inst_parser->setErrorHandler(errReporter);


    // Parse the instance file and catch any exceptions that propogate out
    try
    {
        inst_parser->parse(instanceFile);
    }

    catch (const XMLException& toCatch)
    {
        errStrm << "\nFile not found: '" << instanceFile << "'\n"
                << "Exception message is: \n"
                << toCatch.getMessage() << "\n" << EndLn;
        return false;
    }

	//Report the result

	if (inst_handler->isValid())
		outStrm << "This document is valid." << EndLn;
	else
		outStrm << "This document is invalid." << EndLn;

	// Cleanup
    delete errReporter; 
    delete inst_parser;
    delete inst_handler;

	return true;
}


//////////////////////////////////////////////////////////////////////
//  Reading schemata
//////////////////////////////////////////////////////////////////////

static bool read_schema(DOM_Document& doc, const char* schemaFile, 
						bool useVDOMParser)
{
	//
    //  Instantiate the appropriate DOM parser. 
    //
	NonValidatingDOMParser* parser;
    if (useVDOMParser == true)
        parser = new ValidatingDOMParser;
    else
        parser = new NonValidatingDOMParser;

    //
    //  Attach an error handler to the parser. The parser will call back to
    //  methods of the ErrorHandler if it discovers errors during the course
    //  of parsing the XML document.
    //
	
    ErrorHandler *errReporter = new ErrorReporter();
    parser->setErrorHandler(errReporter);

    //
    //  Parse the XML file, catching any XML exceptions that might propogate
    //  out of it.
    //
    try
    {
        parser->parse(schemaFile);
    }    

    catch (const XMLException& e)
    {
        errStrm << "An error occured during parsing\n   Message: "
                << e.getMessage() << EndLn;
        return false;
    }
	
    doc = parser->getDocument();

    // Cleanup
    delete parser;
    delete errReporter; 

	return true;
}


//////////////////////////////////////////////////////////////////////
// The main function
//////////////////////////////////////////////////////////////////////

int main(int argc, char* argv[])
{
	char*	schemaFile;
	char*	instanceFile;
	bool	debugSchema = false;
	bool	debugInstance = false;
    unsigned int parmInd = 0;
	bool     useVDOMParser     = false;

    try
    {
        XMLPlatformUtils::Initialize();
    }

    catch(const XMLException& toCatch)
    {
        errStrm << "Error during XML4C2 Initialization.\n"
                   "  Exception message is \"%s\"\n"
                << toCatch.getMessage() << EndLn;
        return 1;
    }

   // Check command line and extract arguments.
    if (argc < 3)
    {
        usage();
        return 1;
    }

    // Watch for special case help request
    if (strcmp(argv[1], "-?") == 0)
    {
        usage();
        return 0;
    }

	// See if non validating dom parser configuration is requested.
    for (parmInd = 1; parmInd < (unsigned int)argc; parmInd++)
    {
        // Break out on first parm not starting with a dash
        if (argv[parmInd][0] != '-')
            break;

        if (!strcmp(argv[parmInd], "-v")
        ||  !strcmp(argv[parmInd], "-V"))
        {
            useVDOMParser = true;
        }
        else if (!strcmp(argv[parmInd], "-s")
        ||  !strcmp(argv[parmInd], "-S"))
        {
            debugSchema = true;
        }
        else if (!strcmp(argv[parmInd], "-i")
        ||  !strcmp(argv[parmInd], "-I"))
        {
            debugInstance = true;
        }
         else
        {
            usage();
            return 1;
        }
    }

    schemaFile = argv[parmInd];
    instanceFile = argv[parmInd+1];

	outStrm << "Reading the schema" << EndLn;
    DOM_Document doc;
	if (!(read_schema(doc, schemaFile, useVDOMParser)))
		return 1;

	outStrm << "Creating symbol tables and automata" << EndLn;
	Schema schm(doc, debugSchema);
	try {
		schm.initialize();
	}
	catch (XMLCh* msg) {
		errStrm << msg << EndLn;
		return 1;
	}

	if (debugSchema)
		outStrm << schm;

	outStrm << "Validating the instance" << EndLn;
	if (!(read_instance(schm, instanceFile, debugInstance)))
			return 1;

    return 0;
}

